package com.bc.beam.dataio.srtmDem;

import org.esa.beam.util.io.FileUtils;
import org.esa.beam.util.PropertyMap;
import org.esa.beam.util.Guardian;

import java.io.File;
import java.io.IOException;
import java.io.InputStream;
import java.io.FileInputStream;
import java.util.zip.ZipFile;
import java.util.zip.ZipEntry;
import java.util.Properties;
import java.util.Enumeration;

/**
 * Holds information about a SRTM DEM file.
 * @author Norman Fomferra
 */
public class SrtmDemFileInfo {

    private static final EastingNorthingParser PARSER = new EastingNorthingParser();

    private String _fileName;
    private long _fileSize;
    private float _easting;
    private float _northing;
    private float _pixelSizeX;
    private float _pixelSizeY;
    private int _width;
    private int _height;
    private float _noDataValue;

    private SrtmDemFileInfo() {
    }

    public String getFileName() {
        return _fileName;
    }

    public long getFileSize() {
        return _fileSize;
    }

    public float getEasting() {
        return _easting;
    }

    public float getNorthing() {
        return _northing;
    }

    public float getPixelSizeX() {
        return _pixelSizeX;
    }

    public float getPixelSizeY() {
        return _pixelSizeY;
    }

    public int getWidth() {
        return _width;
    }

    public int getHeight() {
        return _height;
    }

    public float getNoDataValue() {
        return _noDataValue;
    }

    public static SrtmDemFileInfo create(File file) throws IOException {
        String name = file.getName();
        String ext = FileUtils.getExtension(name);
        String baseName = FileUtils.getFilenameWithoutExtension(name);
        String zipExt = "";
        if (ext.equalsIgnoreCase(".zip")) {
            zipExt = ext;
            baseName = FileUtils.getFilenameWithoutExtension(baseName);
        }
        File headerFile = new File(file.getParent(), baseName + ".HDR" + zipExt);
        if (headerFile.exists()) {
            return createFromHeaderFile(headerFile);
        }
        headerFile = new File(file.getParent(), baseName + ".hdr" + zipExt);
        if (headerFile.exists()) {
            return createFromHeaderFile(headerFile);
        }
        return createFromDataFile(file);
    }

    private static SrtmDemFileInfo createFromHeaderFile(File headerFile) throws IOException {
        final SrtmDemFileInfo srtmDemFileInfo = new SrtmDemFileInfo();
        srtmDemFileInfo.setFromHeader(headerFile);
        return srtmDemFileInfo;
    }

    private static SrtmDemFileInfo createFromDataFile(File dataFile) throws IOException {
        final SrtmDemFileInfo srtmDemFileInfo = new SrtmDemFileInfo();
        srtmDemFileInfo.setFromData(dataFile);
        return srtmDemFileInfo;
    }

    private void setFromHeader(File headerFile) throws IOException {
        String ext = FileUtils.getExtension(headerFile.getName());
        if (ext.equalsIgnoreCase(".zip")) {
            String baseName = FileUtils.getFilenameWithoutExtension(headerFile.getName());
            ZipFile zipFile = new ZipFile(headerFile);
            try {
                final ZipEntry zipEntry = getZipEntryIgnoreCase(zipFile, baseName);
                if (zipEntry == null) {
                    throw new IOException("Entry '"+baseName+"' not found in Zip file.");
                }
                final InputStream inputStream = zipFile.getInputStream(zipEntry);
                try {
                    setFromHeader(inputStream);
                } finally {
                    inputStream.close();
                }
            } finally {
                zipFile.close();
            }
        } else {
            final FileInputStream fileInputStream = new FileInputStream(headerFile);
            try {
                setFromHeader(fileInputStream);
            } finally {
                fileInputStream.close();
            }
        }
    }

    private ZipEntry getZipEntryIgnoreCase(ZipFile zipFile, String name) {
        final Enumeration enumeration = zipFile.entries();
        while (enumeration.hasMoreElements()) {
            ZipEntry zipEntry = (ZipEntry) enumeration.nextElement();
            if (zipEntry.getName().equalsIgnoreCase(name)) {
                return zipEntry;
            }
        }
        return null;
    }

    public void setFromHeader(InputStream istream) throws IOException {
        final Properties properties = new Properties();
        properties.load(istream);
        setFromHeader(properties);
    }

    public void setFromHeader(Properties properties) throws IOException {
        PropertyMap propertyMap = new PropertyMap(properties);

        _width = propertyMap.getPropertyInt("NCOLS"); //       e.g. 4800
        _height = propertyMap.getPropertyInt("NROWS");  //     e.g. 4800
        _easting = (float) propertyMap.getPropertyDouble("ULXMAP"); //        e.g. 100.00416
        _northing = (float)propertyMap.getPropertyDouble("ULYMAP"); //        e.g. 89.99583
        _pixelSizeX = (float)propertyMap.getPropertyDouble("XDIM"); //        e.g. 0.0083
        _pixelSizeY = (float)propertyMap.getPropertyDouble("YDIM"); //        e.g. 0.0083
        _noDataValue = (float) propertyMap.getPropertyDouble("NODATA"); //    e.g. -9999

        checkValid(propertyMap, "BYTEORDER", "M");
        checkValid(propertyMap, "LAYOUT", "BIL");
        checkValid(propertyMap, "NBANDS", 1);
        checkValid(propertyMap, "NBITS", 16);
        checkValid(propertyMap, "BANDROWBYTES", 2 * _width);
        checkValid(propertyMap, "TOTALROWBYTES", 2 * _width);
        checkValid(propertyMap, "BANDGAPBYTES", 0);
    }

    private void setFromData(File dataFile) throws IOException {
        String ext = FileUtils.getExtension(dataFile.getName());
        if (ext.equalsIgnoreCase(".zip")) {
            String baseName = FileUtils.getFilenameWithoutExtension(dataFile.getName());
            ZipFile zipFile = new ZipFile(dataFile);
            try {
                final ZipEntry zipEntry = getZipEntryIgnoreCase(zipFile, baseName);
                if (zipEntry == null) {
                    throw new IOException("Entry '"+baseName+"' not found in zip file.");
                }
                setFromData(baseName, zipEntry.getSize());
            } finally {
                zipFile.close();
            }
        } else {
            setFromData(dataFile.getName(), dataFile.length());
        }
    }

    public void setFromData(String fileName, long fileSize) throws IOException {

        _fileName = fileName;
        _fileSize = fileSize;

        final int[] values = parseEastingNorthing(fileName);
        if (values == null) {
            throw new IOException("Illegal file name format: " + fileName);
        }
        _easting =  values[0];
        _northing =  values[1];

        _width = (int) Math.sqrt(fileSize / 2);
        _height = _width;
        if (_width * _height * 2L != fileSize) {
            throw new IOException("Illegal file size: " + fileSize);
        }

        _pixelSizeX = 1.0F / (60.0F * 60.0F);  // 1 arcsecond product
        _pixelSizeY = _pixelSizeX;

        _noDataValue = -9999;
    }


    private static void checkValid(PropertyMap propertyMap, String name, int expectedValue) throws IOException {
        int value = propertyMap.getPropertyInt(name, expectedValue);
        if (value != expectedValue) {
            throw new IOException(getErrorMessage(name, expectedValue + ""));
        }
    }

    private static void checkValid(PropertyMap propertyMap, String name, String expectedValue) throws IOException {
        String value = propertyMap.getPropertyString(name, expectedValue);
        if (!value.equalsIgnoreCase(expectedValue)) {
            throw new IOException(getErrorMessage(name, expectedValue));
        }
    }

    private static String getErrorMessage(String name, String expectedValue) {
        return "Illegal image format: " +
                                "Header entry '" + name + "' must be '"+expectedValue+"'";
    }

    public static int[] parseEastingNorthing(String text) {
        Guardian.assertNotNull("text", text);
        if (text.length() == 0) {
            return null;
        }
        return PARSER.parse(text);
    }

    public static boolean isValidFileSize(long size) {
        if (size > 0 && size % 2 == 0) {
            long w = (long) Math.sqrt(size / 2);
            if (2 * w * w == size) {
                return true;
            }
        }
        return false;
    }

    private static class EastingNorthingParser {
        private String text;
        private int pos;
        private static final char EOT = (char) -1;

        private int[] parse(String text) {
            this.text = text;
            this.pos = -1;
            int[] values = parseImpl();
            this.text = null;
            this.pos = -1;
            return values;
        }

        private int[] parseImpl() {
            int easting = -999;
            int northing = -999;

            char c = nextChar();
            boolean isWest = (c == 'w' || c == 'W');
            boolean isEast = (c == 'e' || c == 'E');
            boolean isNorth = (c == 'n' || c == 'N');
            boolean isSouth = (c == 's' || c == 'S');
            if (!isWest && !isEast && !isNorth && !isSouth) {
                return null;
            }

            c = nextChar();
            if (!Character.isDigit(c)) {
                return null;
            }
            int value = nextNumber(c, isWest || isSouth);
            if (isWest || isEast) {
                easting = value;
            } else {
                northing = value;
            }

            c = nextChar();
            isWest = (c == 'w' || c == 'W');
            isEast = (c == 'e' || c == 'E');
            isNorth = (c == 'n' || c == 'N');
            isSouth = (c == 's' || c == 'S');
            if (!isWest && !isEast && !isNorth && !isSouth) {
                return null;
            }

            c = nextChar();
            if (!Character.isDigit(c)) {
                return null;
            }
            value = nextNumber(c, isWest || isSouth);
            if (isWest || isEast) {
                easting = value;
            } else {
                northing = value;
            }

            if (easting < -180 || easting > +180) {
                return null;
            }
            if (northing < -90 || northing > +90) {
                return null;
            }

            c = nextChar();
            if (c != '.' && c != EOT) {
                return null;
            }

            return new int[]{easting, northing};
        }

        private int nextNumber(char c, boolean neg) {
            int value = 0;
            while (Character.isDigit(c)) {
                value *= 10;
                value += (c - '0');
                c = nextChar();
            }
            pos--;
            value *= (neg ? -1 : +1);
            return value;
        }

        private char nextChar() {
            pos++;
            return pos < text.length() ? text.charAt(pos) : EOT;
        }

        private EastingNorthingParser() {
        }

    }
}
