package com.bc.beam.dataio.srtmDem;

import org.esa.beam.framework.dataio.AbstractProductReader;
import org.esa.beam.framework.dataio.IllegalFileFormatException;
import org.esa.beam.framework.datamodel.*;
import org.esa.beam.framework.dataop.maptransf.MapInfo;
import org.esa.beam.framework.dataop.maptransf.MapProjectionRegistry;
import org.esa.beam.framework.dataop.maptransf.IdentityTransformDescriptor;
import org.esa.beam.framework.dataop.maptransf.Datum;
import org.esa.beam.util.io.FileUtils;
import org.esa.beam.util.SystemUtils;

import javax.imageio.stream.ImageInputStream;
import javax.imageio.stream.FileCacheImageInputStream;
import javax.imageio.stream.FileImageInputStream;
import java.io.IOException;
import java.io.File;
import java.io.InputStream;
import java.util.zip.ZipFile;
import java.util.zip.ZipEntry;

import com.bc.progress.ProgressControllerPool;
import com.bc.progress.ProgressController;

/**
 * A product reader for SRTM DEM data.
 * @author Norman Fomferra
 */
public class SrtmDemReader extends AbstractProductReader {
    private ZipFile _zipFile;
    private ImageInputStream _imageInputStream;
    private SrtmDemFileInfo _fileInfo;
    private Product _product;
    private Band _elevationBand;
    private ProgressController _pc;

    public SrtmDemReader(SrtmDemReaderPlugIn readerPlugIn) {
        super(readerPlugIn);
    }

    /**
     * Provides an implementation of the <code>readProductNodes</code> interface method. Clients implementing this
     * method can be sure that the input object and eventually the subset information has already been set.
     * <p/>
     * <p>This method is called as a last step in the <code>readProductNodes(input, subsetInfo)</code> method.
     *
     * @throws java.io.IOException if an I/O error occurs
     */
    protected Product readProductNodesImpl() throws IOException, IllegalFileFormatException {
        initReader();

        final File dataFile = SrtmDemReaderPlugIn.getInputFile(getInput());
        _fileInfo = SrtmDemFileInfo.create(dataFile);

        final String productName;
        try {
            final String ext = FileUtils.getExtension(dataFile);
            if (ext.equalsIgnoreCase(".zip")) {
                final String entryName = FileUtils.getFilenameWithoutExtension(dataFile.getName());
                _zipFile = new ZipFile(dataFile);
                final ZipEntry entry = _zipFile.getEntry(entryName);
                final InputStream inputStream = _zipFile.getInputStream(entry);
                _imageInputStream = new FileCacheImageInputStream(inputStream, createCacheDir());
                productName = FileUtils.getFilenameWithoutExtension(entryName);
            } else {
                _imageInputStream = new FileImageInputStream(dataFile);
                productName = FileUtils.getFilenameWithoutExtension(dataFile);
            }
        } catch (IOException e) {
            try {
                close();
            } catch (IOException ignored) {
            }
            throw e;
        }

        initProduct(productName);
        return _product;
    }

    /**
     * The template method which is called by the method after an optional spatial subset has been applied to the input
     * parameters.
     * <p/>
     * <p>The destination band, buffer and region parameters are exactly the ones passed to the original  call. Since
     * the <code>destOffsetX</code> and <code>destOffsetY</code> parameters are already taken into acount in the
     * <code>sourceOffsetX</code> and <code>sourceOffsetY</code> parameters, an implementor of this method is free to
     * ignore them.
     *
     * @param sourceOffsetX the absolute X-offset in source raster co-ordinates
     * @param sourceOffsetY the absolute Y-offset in source raster co-ordinates
     * @param sourceWidth   the width of region providing samples to be read given in source raster co-ordinates
     * @param sourceHeight  the height of region providing samples to be read given in source raster co-ordinates
     * @param sourceStepX   the sub-sampling in X direction within the region providing samples to be read
     * @param sourceStepY   the sub-sampling in Y direction within the region providing samples to be read
     * @param destBand      the destination band which identifies the data source from which to read the sample values
     * @param destOffsetX   the X-offset in the band's raster co-ordinates
     * @param destOffsetY   the Y-offset in the band's raster co-ordinates
     * @param destWidth     the width of region to be read given in the band's raster co-ordinates
     * @param destHeight    the height of region to be read given in the band's raster co-ordinates
     * @param destBuffer    the destination buffer which receives the sample values to be read
     * @throws java.io.IOException if an I/O error occurs
     * @see #readBandRasterData
     * @see #getSubsetDef
     */
    protected void readBandRasterDataImpl(int sourceOffsetX,
                                          int sourceOffsetY,
                                          int sourceWidth,
                                          int sourceHeight,
                                          int sourceStepX,
                                          int sourceStepY,
                                          Band destBand,
                                          int destOffsetX,
                                          int destOffsetY,
                                          int destWidth,
                                          int destHeight,
                                          ProductData destBuffer) throws IOException {
        _pc = ProgressControllerPool.getInstance().getProgressController();
        try {
             _pc.fireProcessStarted("Reading SRTM-DEM data...", 0, destHeight - 1);
            final short[] elems = (short[]) destBuffer.getElems();
            readRasterDataImpl(elems, sourceOffsetX, sourceOffsetY, sourceStepX, sourceStepY, destWidth, destHeight);
            maskInt32Extrema(elems);
        } finally {
            _pc.fireProcessEnded();
            _pc = null;
        }
    }


    /**
     * Closes the access to all currently opened resources such as file input streams and all resources of this children
     * directly owned by this reader. Its primary use is to allow the garbage collector to perform a vanilla job.
     * <p/>
     * <p>This method should be called only if it is for sure that this object instance will never be used again. The
     * results of referencing an instance of this class after a call to <code>close()</code> are undefined.
     * <p/>
     * <p>Overrides of this method should always call <code>super.close();</code> after disposing this instance.
     *
     * @throws java.io.IOException if an I/O error occurs
     */
    public void close() throws IOException {
        super.close();
        closeImageInputStream();
        closeZipFile();
        _product = null;
    }

    /**
     * Gets the reader's cache directory, the folder "temp" in the BEAM installation directory.
     * @return the reader's cache directory
     */
    public static File getCacheDir() {
        return new File(SystemUtils.getBeamHomeDir(), "temp");
    }

    private void closeImageInputStream() throws IOException {
        if (_imageInputStream != null) {
            _imageInputStream.close();
            _imageInputStream = null;
        }
    }

    private void closeZipFile() throws IOException {
        if (_zipFile != null) {
            _zipFile.close();
            _zipFile = null;
        }
    }

    private void initReader() {
        _zipFile = null;
        _imageInputStream = null;
        _product = null;
        _fileInfo = null;
    }

    private static File createCacheDir() throws IOException {
        final File cacheDir = getCacheDir();
        if (!cacheDir.exists() && !cacheDir.mkdir()) {
            throw new IOException("Failed to create cache directory '" + cacheDir + "'.");
        }
        return cacheDir;
    }

    private void initProduct(final String productName) {
        final int width = _fileInfo.getWidth();
        final int height = _fileInfo.getHeight();
        _product = new Product(productName, SrtmDemReaderPlugIn.FORMAT_NAME, width, height, this);
        final MapInfo mapInfo = new MapInfo(MapProjectionRegistry.getProjection(IdentityTransformDescriptor.NAME),
                0.5F, 0.5F,
                _fileInfo.getEasting(),
                _fileInfo.getNorthing(),
                _fileInfo.getPixelSizeX(),
                _fileInfo.getPixelSizeY(),
                Datum.WGS_84);
        mapInfo.setSceneWidth(width);
        mapInfo.setSceneHeight(height);
        _product.setGeoCoding(new MapGeoCoding(mapInfo));
        _product.setDescription("SRT-DEM data product");
        _elevationBand = new Band("elevation", ProductData.TYPE_INT16, width, height);
        _elevationBand.setUnit("m");
        _elevationBand.setDescription("SRTM elevation");
        _product.addBand(_elevationBand);
    }


    private void readRasterDataImpl(final short[] elems,
                                    final int sourceOffsetX,
                                    final int sourceOffsetY,
                                    final int sourceStepX,
                                    final int sourceStepY,
                                    final int destWidth,
                                    final int destHeight) throws IOException {
        final int numPixelsTotal = _product.getSceneRasterWidth() * _product.getSceneRasterHeight();
        final int sceneWidth = _product.getSceneRasterWidth();
        if (sourceStepX == 1) {
            int sourceY = sourceOffsetY;
            for (int destY = 0; destY < destHeight; destY++) {

                final long sourcePos = sourceY * sceneWidth + sourceOffsetX;
                checkValidSourcePos(sourcePos, destWidth, numPixelsTotal);
                _imageInputStream.seek(2 * sourcePos);

                final int destPos = destY * destWidth;
                _imageInputStream.readFully(elems, destPos, destWidth);
                sourceY += sourceStepY;
                _pc.fireProcessInProgress(destY);
            }
        } else {
            int sourceY = sourceOffsetY;
            for (int destY = 0; destY < destHeight; destY++) {
                int sourceX = sourceOffsetX;
                for (int destX = 0; destX < destWidth; destX++) {

                    final long sourcePos = sourceY * sceneWidth + sourceX;
                    checkValidSourcePos(sourcePos, destWidth, numPixelsTotal);
                    _imageInputStream.seek(2 * sourcePos);

                    final int destPos = destY * destWidth + destX;
                    elems[destPos] = _imageInputStream.readShort();
                    sourceX += sourceStepX;
                }
                sourceY += sourceStepY;
                _pc.fireProcessInProgress(destY);
            }
        }
    }

    private static void checkValidSourcePos(final long sourcePos, final int destWidth, final int numPixelsTotal) {
        if (sourcePos + destWidth > numPixelsTotal) {
            throw new IllegalStateException("sourcePos + destWidth > numPixelsTotal");
        }
    }

    private void maskInt32Extrema(final short[] elems) {
        final short noDataValue = (short) _fileInfo.getNoDataValue();
        for (int i = 0; i < elems.length; i++) {
            if (elems[i] == Short.MIN_VALUE || elems[i] == Short.MAX_VALUE) {
                elems[i] = noDataValue;
            }
        }
    }
}
