package com.bc.beam.visat;

import com.bc.beam.dataio.srtmDem.SrtmDemReader;
import com.bc.beam.dataio.srtmDem.SrtmDemReaderPlugIn;
import org.esa.beam.framework.dataio.ProductIOPlugInManager;
import org.esa.beam.framework.ui.io.BeamFileFilter;
import org.esa.beam.visat.AbstractProductImportPlugIn;
import org.esa.beam.visat.VisatApp;

import java.io.File;

/**
 * A VISAT plugin which allows for reading SRTM-DEM data.
 *
 * @author Norman Fomferra
 */
public class SrtmDemReaderVPI extends AbstractProductImportPlugIn {
    final static BeamFileFilter FILE_FILTER = new FixedBeamFileFilter();

    public SrtmDemReaderVPI() {
        super("importSrtmDem", FILE_FILTER, false);
    }

    /**
     * Called by VISAT after the plug-in instance has been registered in VISAT's plug-in manager.
     *
     * @param visatApp a reference to the VISAT application instance.
     * @see org.esa.beam.visat.VisatPlugIn#initPlugIn(org.esa.beam.visat.VisatApp)
     */
    public void initPlugIn(VisatApp visatApp) {
        super.initPlugIn(visatApp);
        ProductIOPlugInManager.getInstance().addReaderPlugIn(new SrtmDemReaderPlugIn());
        getCommand().setParent("import");
        getCommand().setMnemonic('D');
        getCommand().setText("Import SRTM-DEM Product (or Subset)...");
        getCommand().setShortDescription("Import a SRTM-DEM data product (subset).");
        getCommand().setLongDescription("Import a SRTM-DEM data product (subset).");

        Runtime.getRuntime().addShutdownHook(new Thread() {
            public void run() {
                clearCacheDir();
            }
        });
    }

    private void clearCacheDir() {
        final File cacheDir = SrtmDemReader.getCacheDir();
        if (cacheDir.exists()) {
            final File[] files = cacheDir.listFiles();
            if (files.length > 0) {
                VisatApp.getApp().getLogger().info("Clearing cache directory " + cacheDir.getPath());
                for (int i = 0; i < files.length; i++) {
                    VisatApp.getApp().getLogger().info("Deleting " + files[i]);
                    files[i].delete();
                }
            }
        }
    }

    /**
     * This class fixes a tiny problem of the BeamFileFilter as of BEAM 3.1.
     * The 3.2 BeamFileFilter could not handle extensions such as ".dem.zip"
     * correctly.
     */
    private static class FixedBeamFileFilter extends BeamFileFilter {

        public FixedBeamFileFilter() {
            super(SrtmDemReaderPlugIn.FORMAT_NAME, SrtmDemReaderPlugIn.FILE_EXTENSIONS, SrtmDemReaderPlugIn.DESCRIPTION);
        }

        /**
         * Overrides the base lass implementation in order to compare
         * the extensions correctly.
         *
         * @param filename
         * @return
         */
        public boolean checkExtension(String filename) {
            final String filenameLC = filename.toLowerCase();
            final String[] extensions = getExtensions();
            for (int i = 0; i < extensions.length; i++) {
                String extension = extensions[i];
                if (filenameLC.endsWith(extension) &&
                        !filenameLC.equals(extension)) {
                    return true;
                }
            }
            return false;
        }
    }
}
